#!/usr/bin/env python

# @uthor Benoit Neviani (benoit.neviani@viavisolutions.com)
# version v.2.0

# import
import requests
import json
import os
import stat
import socket
import argparse
import logging
import datetime
import uuid
import time
import calendar
import urllib3

# argurments check
parser = argparse.ArgumentParser()
parser.add_argument("-v","--version", action='version', version='%(prog)s : version v.2.0')
parser.add_argument("-define", action='store_const', const='define', dest='funct', help='define variables', metavar='')
parser.add_argument("-now", action='store_const', const='now', dest='funct', help='retrieve data in real-time [from the last minute]')
parser.add_argument("-past", action='store_const', const='past', dest='funct', help='retrieve data in the past [select start/end]')
parser.add_argument("-auto", action='store_const', const='auto', dest='funct', help='retrieve data automatically [time periods without data in apex.nolog]')

parser.add_argument("--debug", help='debug level', action='store_const', const='debug', metavar='')
whattodo = parser.parse_args()
args, leftovers = parser.parse_known_args()

if args.debug is not None :
     logging.basicConfig(level=logging.DEBUG)
     logging.getLogger("requests").setLevel(logging.DEBUG)
else :
     logging.basicConfig(level=logging.INFO)
     logging.getLogger("requests").setLevel(logging.WARNING)
logger = logging.getLogger()
formatter = logging.Formatter('%(asctime)s :: %(levelname)s :: %(message)s')
handler = logging.FileHandler('apex.log')
logger.addHandler(handler)
handler.setFormatter(formatter)

# functions
##variables
def writevar(name,value):
    write = {name:value}
    try:
        with open('var.json', 'a') as var:
            json.dump(write, var)
            var.write(os.linesep)
            var.close()
    except IOError as e:
            logger.info("impossible to write var.json, please check permission/disk space...")
            exit(1)
def readvar(name,file):
    data = []
    with open(file, 'r') as var:
        for line in var:
            data.append(json.loads(line))
        for key in data:
            if name in key:
                return key[name]

def defvariables():
    random = uuid.uuid4().hex
    uniqid = random[:8] + "-"  + random[9:13] + "-"  + random[13:17] + "-"  + random[17:21] + "-"  + random[20:]
    if calendar.timegm(time.gmtime()) - calendar.timegm(time.localtime()) != 0:
        print("Your system is in UTC?")
    logpath = input("Full path to write JSON data: ")
    while logpath is '':
        logpath = input("Please enter Full path to write JSON data: ")
    if logpath.endswith("/"):
        writevar("logpath",logpath)
    else:
        writevar("logpath",logpath + "/")
    writevar("timeshift","5")
    writevar("baseURL","https://")
    writevar("sessionURL","/oa/api/session")
    writevar("dashboardsURL","/oa/api/dashboards")
    writevar("businessgroupURL","/oa/api/businessGroups")
    writevar("widgetsURL","/oa/api/widgets")
    writevar("dashboardresultsURL","/oa/api/dashboardResults")
    ipAddress = input("Apex IP Address/DNS name: ")
    while ipAddress is '':
        ipAddress = input("Please enter Apex IP Address/DNS name: ")
    writevar("ipAddress",ipAddress)
    username = input("Apex User: ")
    while username is '':
        username = input("Please enter Apex User: ")
    writevar("username",username)
    password = input("Apex Password: ")
    while password is '':
        password = input("Please enter Apex Password: ")
    writevar("password",password)
    bg = input("Apex EXISTING Business Group: ")
    while bg is '':
        bg = input("Please enter Apex EXISTING Business Group: ")
    writevar("bg",bg)
    dashboard = input("Apex Dashboard: ")
    while dashboard is '':
        dashboard = input("Please enter Apex Dashboard: ")
    writevar("dashboard",dashboard)
    widget = input("Apex Widget: ")
    while widget is '':
        widget = input("Please enter Apex Widget: ")
    writevar("widget",widget)
    writevar("id",uniqid)
    userid = getUserId()
    writevar("UserId",userid)
    logger.info("var.json created successfully, ids to follow...")

def testvariables():
    sessionCookie = login()
    drag_and_drop = 'no'
    if not getBusinessGroupId(sessionCookie):
        logger.info("There is no Business Group named " + readvar("bg","var.json") + " in Apex please check the Business Group name in Apex !")
        os.remove('var.json')
        exit(1)
    if not getDashboardId(sessionCookie):
        logger.info("Dashboard doesnt exist, creating the Dashboard in Apex now")
        create_dashboard(sessionCookie, readvar("UserId","var.json"))
        drag_and_drop = "yes"
    if not getWidgetId(sessionCookie):
        logger.info("Widget doesnt exist, creating the Widget in Apex now")
        create_widget(sessionCookie, readvar("UserId","var.json"))
    writevar("BusinessGroupId",getBusinessGroupId(sessionCookie))
    writevar("DashboardId",getDashboardId(sessionCookie))
    writevar("DashboardEtag",getDashboardEtag(sessionCookie))
    writevar("WidgetId",getWidgetId(sessionCookie))
    if drag_and_drop is "yes":
        logger.info("Moving Widget to the Dashboard in Apex now")
        create_widget_on_dashboard(sessionCookie, readvar("UserId","var.json"))
    logout(sessionCookie)

## http methods
def getRequest(sessionCookie, url):
    Cookie = "SessionApexId="+sessionCookie
    headers = {'Content-Type':'Application/json','Accept':'Application/json','Cookie':Cookie}
    request = requests.get(url, headers = headers, verify = False)
    logger.debug("HTTP GET is " + str(request) + "\n")
    return request

def postRequest(sessionCookie, url, data):
    Cookie = "SessionApexId="+sessionCookie
    headers = {'Content-Type':'Application/json','Accept':'Application/json','Cookie':Cookie}
    request = requests.post(url, headers = headers, data = data, verify = False)
    logger.debug("HTTP POST is " + str(request) + "\n")
    return request

def putRequest(sessionCookie, url, data):
    Cookie = "SessionApexId="+sessionCookie
    headers = {'Content-Type':'Application/json','Accept':'Application/json','Cookie':Cookie}
    request = requests.put(url, headers = headers, data = data, verify = False)
    logger.debug("HTTP PUT is " + str(request) + "\n")
    return request

def deleteRequest(sessionCookie, url):
    Cookie = "SessionApexId="+sessionCookie
    headers = {'Content-Type':'Application/json','Accept':'Application/json','Cookie':Cookie}
    request = requests.delete(url, headers = headers, verify = False)
    logger.debug("HTTP DELETE is " + str(request) + "\n")
    return request

## sessions management
def login():
    url = readvar("baseURL","var.json") + readvar("ipAddress","var.json") + readvar("sessionURL","var.json")
    data = '{"session": {"username": "'+readvar("username","var.json")+'","password": "'+readvar("password","var.json")+'"}}'
    logger.debug("Apex data request is " + str(data) + "\n")
    response = postRequest('nothing', url, data)
    logger.debug("Apex login data response is " + str(response.text) + "\n")
    if (response.ok):
        logger.info("Login successfully")
        cookie = response.cookies['SessionApexId']
        logger.debug("SessionApexId is " + str(cookie) + "\n")
        return cookie
    else:
        logger.info("Login failed")
        now = datetime.datetime.utcnow().replace(second=0,microsecond=0)
        now_less_timeshift = now - datetime.timedelta(minutes=int(readvar("timeshift","var.json")))
        write_nolog_file(iso(now_less_timeshift))
        exit(1)

def logout(sessionCookie):
    url = readvar("baseURL","var.json") + readvar("ipAddress","var.json") + readvar("sessionURL","var.json")
    response = deleteRequest(sessionCookie, url)
    logger.debug("Apex logout data response is " + str(response.text) + "\n")
    if (response.ok):
            logger.info("Logout successfully")
    else:
            logger.info("Logout Failed")
            exit(1)

## monitoring
def monitoring():
    logger.info("*****Start script*****")
    ping = os.system("ping -c 1 -w2 " + readvar("ipAddress","var.json") + " 2>&1 >/dev/null")
    if ping == 0:
        logger.info("Apex is up")
    else:
        logger.info("Apex is down")
        now = datetime.datetime.utcnow().replace(second=0,microsecond=0)
        now_less_timeshift = now - datetime.timedelta(minutes=int(readvar("timeshift","var.json")))
        write_nolog_file(iso(now_less_timeshift))
        exit(1)
    s = socket.socket()
    try:
        s.connect((readvar("ipAddress","var.json"), 443))
    except Exception as connection:
        logger.info("Apex connection refused")
        now = datetime.datetime.utcnow().replace(second=0,microsecond=0)
        now_less_timeshift = now - datetime.timedelta(minutes=int(readvar("timeshift","var.json")))
        write_nolog_file(iso(now_less_timeshift))
        exit(1)
    finally:
        s.close()

## retrieve Ids
def getBusinessGroupId(sessionCookie):
    url = readvar("baseURL","var.json") + readvar("ipAddress","var.json") + readvar("businessgroupURL","var.json")
    response = getRequest(sessionCookie, url)
    json_response = json.loads(response.text)
    logger.debug("Apex business group data response is " + str(response.text) + "\n")
    bg_id = ''
    for key in json_response['data']['businessgroups']:
        if key['name'] == readvar("bg","var.json"):
            bg_id = key['id']
            logger.debug("Apex business group " + readvar("bg","var.json") + " id is " + str(bg_id) + "\n")
            return bg_id
    if bg_id == '':
        logger.info("Business Group id not found")

def getDashboardId(sessionCookie):
    url = readvar("baseURL","var.json") + readvar("ipAddress","var.json") + readvar("dashboardsURL","var.json")
    response = getRequest(sessionCookie, url)
    json_response = json.loads(response.text)
    logger.debug("Apex dashboards data response is " + str(response.text) + "\n")
    dashboard_id = ''
    for key in json_response['data']['dashboards']:
        if key['title'] == readvar("dashboard","var.json"):
            dashboard_id = key['id']
            logger.debug("Apex dashboard " + readvar("dashboard","var.json") + " id is " + str(dashboard_id) + "\n")
            return dashboard_id
    if dashboard_id == '':
        logger.info("Dashboard id not found")

def getDashboardEtag(sessionCookie):
    url = readvar("baseURL","var.json") + readvar("ipAddress","var.json") + readvar("dashboardsURL","var.json")
    response = getRequest(sessionCookie, url)
    json_response = json.loads(response.text)
    logger.debug("Apex dashboards data response is " + str(response.text) + "\n")
    dashboard_etag = ''
    for key in json_response['data']['dashboards']:
        if key['title'] == readvar("dashboard","var.json"):
            dashboard_etag = key['etag']
            logger.debug("Apex dashboard " + readvar("dashboard","var.json") + " etag is " + str(dashboard_etag) + "\n")
            return dashboard_etag
    if dashboard_etag == '':
        logger.info("Dashboard etag not found")

def getWidgetId(sessionCookie):
    url = readvar("baseURL","var.json") + readvar("ipAddress","var.json") + readvar("widgetsURL","var.json")
    response = getRequest(sessionCookie, url)
    json_response = json.loads(response.text)
    logger.debug("Apex widget data response is " + str(response.text) + "\n")
    widget_id = ''
    for key in json_response['data']['widgets']:
        if key['title'] == readvar("widget","var.json"):
            widget_id = key['id']
            logger.debug("Apex widget " + readvar("widget","var.json") + " id is " + str(widget_id) + "\n")
            return widget_id
    if widget_id == '':
        logger.info("Widget id not found")

def getUserId():
    url = readvar("baseURL","var.json") + readvar("ipAddress","var.json") + readvar("sessionURL","var.json")
    data = '{"session": {"username": "'+readvar("username","var.json")+'","password": "'+readvar("password","var.json")+'"}}'
    logger.debug("Apex data request is " + str(data) + "\n")
    response = postRequest('nothing', url, data)
    json_response = json.loads(response.text)
    logger.debug("Apex login data response is " + str(response.text) + "\n")
    if (response.ok):
        logger.info("Login successfully")
        sessionCookie = response.cookies['SessionApexId']
        userid = json_response['data']['LoginSession']['user']['id']
        logger.debug("UserId is " + str(userid) + "\n")
        deleteRequest(sessionCookie, url)
        return userid
    else:
        logger.info("Login failed")
        now = datetime.datetime.utcnow().replace(second=0,microsecond=0)
        now_less_timeshift = now - datetime.timedelta(minutes=int(readvar("timeshift","var.json")))
        write_nolog_file(iso(now_less_timeshift))
        exit(1)

## create functions
def create_dashboard(sessionCookie, user_id):
    url = readvar("baseURL","var.json") + readvar("ipAddress","var.json") + readvar("dashboardsURL","var.json")
    data = '{"dashboard": {"owner": {"ownerType": "USER","userId": "'+readvar("UserId","var.json")+'"},"title": "'+readvar("dashboard","var.json")+'","description": "Dashboard Generated by script","layout":{"type": "NAMED","name":"1"}}}'
    logger.debug("Apex data request data is " + str(data) + "\n")
    response = postRequest(sessionCookie, url, data)
    json_response = json.loads(response.text)
    logger.debug("Apex dashboards data response is " + str(response.text) + "\n")

def create_widget(sessionCookie, user_id):
    url = readvar("baseURL","var.json") + readvar("ipAddress","var.json") + readvar("widgetsURL","var.json")
    print("*****Please enter KEY field(s) number(s) separated by space (sorted by the first KEY field)*****")
    print("Example: 34 53 61 62 produces a Widget Table: | IP Address  | Bytes Total | Bits/sec Rx | Bits/sec Tx")
    key_fields_list_selection = input("Do you want to see available KEY fields? (y/n) ")
    while key_fields_list_selection not in ['y', 'n']:
        key_fields_list_selection = input("Do you want to see available KEY fields? (y/n) ")
    if key_fields_list_selection is "y":
        key_fields_list = open('keyfields.json', 'r')
        print (key_fields_list.read())
    key_fields_input = input("Please enter your KEY fields selection: ")
    data_fields_list_selection = input("Do you want to see available DATA fields? (y/n) ")
    while data_fields_list_selection not in ['y', 'n']:
        data_fields_list_selection = input("Do you want to see available DATA fields? (y/n) ")
    if data_fields_list_selection is "y":
        data_fields_list = open('datafields.json', 'r')
        print (data_fields_list.read())
    data_fields_input = input("Please enter your DATA fields selection: ")
    key_fields_table = key_fields_input.split()
    data_fields_table = data_fields_input.split()
    if key_fields_input == '':
        data1 = '{"widget": {"owner": {"ownerType": "USER","userId": "'+readvar("UserId","var.json")+'"},"options": {"numberFormat": "METRIC"}, "fields": [{"fieldId": '+data_fields_table.pop(0)+',"flags": {"key": true,"display": true},"sortOrder": "DESCENDING","sortIndex":1}'
    else:
        data1 = '{"widget": {"owner": {"ownerType": "USER","userId": "'+readvar("UserId","var.json")+'"},"options": {"numberFormat": "METRIC"}, "fields": [{"fieldId": '+key_fields_table.pop(0)+',"flags": {"key": true,"display": true},"sortOrder": "DESCENDING","sortIndex":1}'
    data2 = ''
    data3 = ''
    for field in key_fields_table:
        keyfield = ',{"fieldId" : '+field+',"flags": {"key": true,"display": true}}'
        data2 += keyfield
    for field in data_fields_table:
        datafield = ',{"fieldId" : '+field+',"flags": {"display": true}}'
        data3 += datafield
    data4 = '],"title": "'+readvar("widget","var.json")+'","widgetType": "TABLE","category": "INTERVAL","maxItems":30000}}'
    if key_fields_input == '':
         data = data1 + data3 + data4
    if data_fields_input == '':
         data = data1 + data2 + data4
    else:
         data = data1 + data2 + data3 + data4
    logger.debug("Apex data request is " + str(data) + "\n")
    response = postRequest(sessionCookie, url, data)
    json_response = json.loads(response.text)
    logger.debug("Apex widget data response is " + str(response.text) + "\n")

def create_widget_on_dashboard(sessionCookie, user_id):
    url = readvar("baseURL","var.json") + readvar("ipAddress","var.json") + readvar("dashboardsURL","var.json") + "/" + readvar("DashboardId","var.json")
    data = '{"dashboard": {"id": "'+readvar("DashboardId","var.json")+'","owner": {"ownerType": "USER","userId": "'+readvar("UserId","var.json")+'"},"title": "'+readvar("dashboard","var.json")+'","description": "Dashboard Generated by script","layout":{"type": "NAMED","name":"1"},"children":[{"elementId": "'+readvar("id","var.json")+'"}],"elements": [{"id": "'+readvar("id","var.json")+'","elementType":"REFERENCE","containerIndex":0,"widgetId": "'+readvar("WidgetId","var.json")+'"}],"etag": "'+readvar("DashboardEtag","var.json")+'","description": "","hideEmptyWidgets": false}}'
    logger.debug("Apex data request is " + str(data) + "\n")
    response = putRequest(sessionCookie, url, data)
    json_response = json.loads(response.text)
    logger.debug("Apex widget to dashboard data response is " + str(response.text) + "\n")

## data functions
def retrieve_data(sessionCookie, time_start, time_end):
    url = readvar("baseURL","var.json") + readvar("ipAddress","var.json") + readvar("dashboardresultsURL","var.json")
    data = '{ "params": {"dashboardId": "'+readvar("DashboardId","var.json")+'","businessGroupId": "'+readvar("BusinessGroupId","var.json")+'","params": {"filter": "","startTime": "'+time_start+'","endTime": "'+time_end+'"}}}'
    logger.debug("Apex data request is " + str(data) + "\n")
    response = postRequest(sessionCookie, url, data)
    json_response = json.loads(response.text)
    logger.debug("Apex retrieve data response is " + str(response.text) + "\n")
    if not (json_response['data']['results']['results'][0]['data']['series']):
        logger.info("Apex retrieve data is blank for period " + str(time_start))
        return 0
    else:
        write_log_file(json_response,time_start)
        return 1

def write_log_file(json_response,time):
    try:
        with open(readvar("logpath","var.json") + time + '.json', 'a') as log:
            series = json_response['data']['results']['results'][0]['data']['series']
            timestamp = json_response['data']['results']['results'][0]['data']['intervalData']['startTime']
            items = 0
            for s in series:
                dataField = s['fieldData']
                keyField = s['legend']
                all = {"timestamp" : timestamp}
                status = 'VALID'
                for l in keyField:
                    keyField = readvar(str(l['fieldId']),"fields.json")
                    keyFieldStatus = l['data']['status']
                    if 'INVALID' in keyFieldStatus:
                        status='INVALID'
                    keyFieldName = l['data']['string']
                    all.update({keyField:keyFieldName})
                for d in dataField:
                    dataFieldStatus = d['data'][0]['status']
                    if 'INVALID' in dataFieldStatus:
                        status='INVALID'
                    dataField = readvar(str(d['fieldId']),"fields.json")
                    if 'unsigned' in d['data'][0]:
                        dataFieldResult = d['data'][0]['unsigned']
                    elif 'float' in d['data'][0]:
                        dataFieldResult = d['data'][0]['float']
                    elif 'time' in d['data'][0]:
                        dataFieldResult = d['data'][0]['time']
                    all.update({dataField:dataFieldResult})
                items += 1
                all.update({"Status":status})
                json.dump(all,log)
                log.write("\n")
            log.close()
            logger.info("number of individual items: " + str(items))
    except IOError as e:
           logger.info("impossible to write data, please check permission/disk space...")
           write_nolog_file(time)

def write_nolog_file(time):
    try:
        with open('apex.nolog', 'a') as nolog:
            nolog.write(time + "\n")
            nolog.close()
    except IOError as e:
           logger.info("apex.nolog is not accessible or impossible to write, please check...")
##time
def iso(time):
    return time.isoformat() + ".00Z"

## execution scheme
try:
    with open('var.json') as file:
        logger.debug("read var.json")
except IOError as e:
    if whattodo.funct is 'define':
        logger.debug("var.json doesn't exist or not accessible trying to create a fresh var.json file now...")
    else:
        logger.info("var.json doesn't exist or not accessible please check permission or run 'apex_a.py -define'")
        exit(1)
if whattodo.funct is None:
    print ("Nothing to do")
    print ("apex_a.py -h for help")
elif whattodo.funct is 'define':
    try:
        os.remove('var.json')
    except OSError:
        pass
    defvariables()
    monitoring()
    testvariables()
    print("*****To automatize data retrieving, please see README cron section*****")

elif 'now' in whattodo.funct:
    monitoring()
    sessionCookie = login()
    now = datetime.datetime.utcnow().replace(second=0,microsecond=0)
    now_less_timeshift = now - datetime.timedelta(minutes=int(readvar("timeshift","var.json")))
    now_less_timeshift_plus_one = now - datetime.timedelta(minutes=int(readvar("timeshift","var.json"))) + datetime.timedelta(minutes=1)
    if retrieve_data(sessionCookie, iso(now_less_timeshift), iso(now_less_timeshift_plus_one)):
        pass
    else:
        write_nolog_file(iso(now_less_timeshift))
    logout(sessionCookie)
elif 'past' in whattodo.funct:
    monitoring()
    sessionCookie = login()
    print ("Time format: YYYY-MM-DD HH:MM Example: 2016-06-29 19:12")
    start = datetime.datetime.strptime(input("Start time: ") + ":00",'%Y-%m-%d %H:%M:%S')
    end = datetime.datetime.strptime(input("End time: ") + ":00",'%Y-%m-%d %H:%M:%S')
    while start != end:
        start_plus_one = start + datetime.timedelta(minutes=1)
        retrieve_data(sessionCookie, iso(start), iso(start_plus_one))
        print(start)
        start = start + datetime.timedelta(minutes=1)
    logout(sessionCookie)
elif 'auto' in whattodo.funct:
    monitoring()
    sessionCookie = login()
    try:
        with open('apex.nolog') as input:
            with open('newapex.nolog', "w") as output:
                for line in input:
                    start = datetime.datetime.strptime(line.strip('\n'), "%Y-%m-%dT%H:%M:%S.00Z")
                    end = start + datetime.timedelta(minutes=1)
                    if not retrieve_data(sessionCookie, iso(start), iso(end)):
                        output.write(line)
                    else :
                        print(start)
        os.remove('apex.nolog')
        os.rename('newapex.nolog','apex.nolog')
        logout(sessionCookie)
    except IOError as e:
        print("I don't find apex.nolog or I have a permission issue, trying to create it now")
        open('apex.nolog', "w").close()
        print("Please send your command again")
    logout(sessionCookie)
